"""
.. WARNING:: Except for :exc:`CoolqliteError` and :exc:`CoolqliteExceptionGroup`,
   these are unstable and may change.
"""

from coolqlite._common import (
    CoolqliteError,
    CoolqliteExceptionGroup,
    MissingNamedToSqlFnError,
    MissingTypedToSqlFnError,
    MissingToSqlFnError,
    ToSqlFnSetupError,
    ToSqlTypeError,
    FromSqlTypeError,
)
from coolqlite.transactions import (
    IncorrectTransactionControlError,
)

from coolqlite.query import (
    ParamMissingToSqlFnError,
    ToSqlFnFailure,
    QueryBuilderErrorGroup,
)

from coolqlite._results import ResultCountError

__all__ = [
    "CoolqliteError",
    "CoolqliteExceptionGroup",
    "MissingNamedToSqlFnError",
    "MissingTypedToSqlFnError",
    "MissingToSqlFnError",
    "ToSqlFnSetupError",
    "ToSqlTypeError",
    "FromSqlTypeError",
    "IncorrectTransactionControlError",
    "ParamMissingToSqlFnError",
    "ToSqlFnFailure",
    "QueryBuilderErrorGroup",
    "ResultCountError",
]



"""
Transaction control through context managers.

We use the savepoint API because it's the most flexible.
We don't use standard ``BEGIN..COMMIT/ROLLBACK`` because
managing the interactions between that and savepoints would be annoying.

Basic Usage
===========

At its simplest, it works the same as how the :mod:`stdlib module <sqlite3>`
uses connections for transactions, but instead with a separate object.

For the following examples, let's use a simple setup:

.. literalinclude:: /../examples/savepoints_basic.py
    :start-after: # region: setup
    :end-before: # endregion setup

Using a :class:`~coolqlite.transactions.Savepoint` context manager
from :meth:`~coolqlite.Connection.with_savepoint`
will "commit" the data if it exits without error.

.. literalinclude:: /../examples/savepoints_basic.py
    :start-after: # region: success
    :end-before: # endregion success

If there's an exception, the transaction will
be rolled back.

.. literalinclude:: /../examples/savepoints_basic.py
    :start-after: # region: errors
    :end-before: # endregion errors
"""

from enum import Enum
import typing
import sqlite3 as sql3
from typing import TYPE_CHECKING, Any, Literal, Self

from coolqlite._common import _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK, CoolqliteError


if TYPE_CHECKING:
    from coolqlite._connection import Connection


class _SavepointStatus(Enum):
    JUST_BORN = "just born"
    ACTIVE = "active"
    DEAD = "dead"


class IncorrectTransactionControlError(CoolqliteError):
    """Something incorrect happened when using savepoints.
    This should only happen if using the savepoint context manager
    incorrectly, or otherwise doing something that would break
    a savepoint correctness guarantee.
    """

    pass


class Savepoint:
    def __init__(
        self,
        private,
        id: str,
        connec: Connection[Any],
        parent: Savepoint | None,
    ):
        assert private is _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK
        self._id = id
        self._status: _SavepointStatus = _SavepointStatus.JUST_BORN
        self._connec = connec
        self._parent = parent

    def __enter__(self) -> Self:
        match self._status:
            case _SavepointStatus.JUST_BORN:
                self._status = _SavepointStatus.ACTIVE
                return self
            case _SavepointStatus.ACTIVE:
                raise IncorrectTransactionControlError(
                    "tried to enter the same savepoint more than once"
                )
            case _SavepointStatus.DEAD:
                raise IncorrectTransactionControlError(
                    "tried to enter an already-used savepoint"
                )

    # TODO: with this simplified one, when
    # do we actually need to track liveliness?

    def __exit__(self, _exc_type, exc_val: BaseException | None, _tb) -> Literal[False]:
        # I guess we also might want an exception for
        # "I know we're invalidated just get me out of here".
        dead = self._status is _SavepointStatus.DEAD
        match (dead, exc_val):
            case (True, _):
                return False
            case (False, None):
                self._do_release()
                return False
            case (False, sql3.IntegrityError()) if (
                not self._connec._connec.in_transaction
            ):
                # this was a ON CONFLICT ROLLBACK,
                # so all savepoints are now invalidated.
                self._connec._invalidate_all_savepoints_due_to_on_conflict_rollback()
                return False
            case (False, _):
                # either a third-party exception
                # or a TransactionControl that wasn't for us.
                # in either case, rollback-to-before
                # but keep it going.
                self._do_rollback_before()
                return False
            case unreachable:  # pragma: no cover
                typing.assert_never(unreachable)

    def _do_release(self):
        self._connec._release(self._id)
        self._status = _SavepointStatus.DEAD

    def _do_rollback_before(self):
        self._connec._rollback_to_before(self._id)
        self._status = _SavepointStatus.DEAD



"""A cool sqlite library."""

from coolqlite._connection import (
    Connection,
    Row,
    ToSqlFnRegistry,
    connect,
)
from coolqlite._results import Results, as_results, QueryResults, first_col
from coolqlite._common import RowFactory, SqliteType
from coolqlite.errors import (
    CoolqliteError,
    CoolqliteExceptionGroup,
    ResultCountError,
)


__all__ = [
    "connect",
    "Connection",
    "Results",
    "as_results",
    "QueryResults",
    "first_col",
    "Row",
    "CoolqliteError",
    "CoolqliteExceptionGroup",
    "ToSqlFnRegistry",
    "RowFactory",
    "SqliteType",
    "ResultCountError",
]



"""
Simple functions for converting from strings to datetype DateTimes.

These all take an optional and ignored second argument
so they can be used as cattrs structure hooks.
"""

import datetime as dt
import re
from typing import cast
import datetype
from datetype import (
    AwareDateTime,
    DateTime,
    NaiveDateTime,
    Date,
    AwareTime,
    NaiveTime,
    Time,
)


# @overload
# def temporal_from_iso(data: str, type_: type[AwareDateTime]) -> AwareDateTime: ...
# @overload
# def temporal_from_iso(data: str, type_: type[DateTime]) -> DateTime: ...
# @overload
# def temporal_from_iso(data: str, type_: type[NaiveDateTime]) -> NaiveDateTime: ...
# @overload
# def temporal_from_iso(data: str, type_: type[Date]) -> Date: ...
# @overload
# def temporal_from_iso(data: str, type_: type[AwareTime]) -> AwareTime: ...
# @overload
# def temporal_from_iso(data: str, type_: type[NaiveTime]) -> NaiveTime: ...
# @overload
# def temporal_from_iso(data: str, type_: type[Time]) -> Time: ...

# def temporal_from_iso[T](data: str, type_: type[T]) -> T:
#     raise NotImplementedError

# TODO: configurable datetime-allowed-from-date?


def dt_from_iso(data: str, _ignored=None, /) -> DateTime:
    "Deserialize a DateTime from an iso-formatted string."

    if not isinstance(data, str):
        raise TypeError(f"Need string not {type(data)} for datetime")

    match re.split(r"[ T]", data):
        case [_date, _time]:
            return datetype.fromisoformat(data)
        case _:
            err = ValueError(
                f"expected to two parts for datetime, date and time, but didn't look right: {data}"
            )
            err.add_note(
                "plain dates are not accepted as datetimes, "
                "add your own struct hook to change this behavior"
            )
            raise err


def naive_dt_from_iso(data: str, _ignored=None, /) -> NaiveDateTime:
    """Deserialize a NaiveDateTime from an iso-formatted string.

    Raises a :exc:`TypeError` if the string had timezone information.
    """
    return datetype.naive(cast(dt.datetime, dt_from_iso(data)))


def aware_dt_from_iso(data: str, _ignored=None, /) -> AwareDateTime:
    """Deserialize an AwareDateTime from an iso-formatted string.

    Raises a :exc:`TypeError` if the string was missing timezone information.
    """
    return datetype.aware(cast(dt.datetime, dt_from_iso(data)))


def time_from_iso(data: str, _ignored=None, /) -> Time:
    if not isinstance(data, str):
        raise TypeError(f"Need string not {type(data)} for time")

    return cast(Time, dt.time.fromisoformat(data))


def aware_time_from_iso(data: str, _ignored=True, /) -> AwareTime:
    return datetype.aware(cast(dt.time, time_from_iso(data)))


# TODO: return type said aware before but mypy didn't complain?


def naive_time_from_iso(data: str, _ignored=True, /) -> NaiveTime:
    return datetype.naive(cast(dt.time, time_from_iso(data)))


def date_from_iso(data: str, _ignored=None, /) -> Date:
    if not isinstance(data, str):
        raise TypeError(f"Need string not {type(data)} for date")

    return cast(Date, dt.date.fromisoformat(data))



"""
Query processing.

You probably won't need to use this module
if you're just using :py:meth:`coolqlite.Connection.query`,
but it may be handy if you want to debug your queries or conversions.

Query proccessing happens in two steps,
arbitrarily (but alphabetically!) named "analyze" and "build".

1. Analyze takes a :class:`string.templatelib.Template` (t-string)
   and turns it into an :class:`AnalyzedQuery`,
   which holds the sql query string and named parameters.
   The values of those parameters haven't been converted yet though.
2. Build takes that :class:`AnalyzedQuery` and turns it into a
   :class:`BuiltQuery`, which has run all conversions.
   This step can fail, but it will collect all errors into
   a :class:`CoolqliteExceptionGroup` so you can see each one.

.. warning:: These names will likely change,
   especially if we eventually get an actual query builder.

   These mostly exist for you to debug your queries and conversions;
   it should not be a stable API to depend upon for now.
"""

from string.templatelib import Interpolation, Template
from coolqlite.to_sql import ToSqlFnRegistry
from coolqlite._common import (
    MissingNamedToSqlFnError,
    MissingTypedToSqlFnError,
)
from coolqlite.query._common import (
    AnalyzedQuery,
    BuiltQuery,
    ConvertedParam,
    Param,
    ParamMissingToSqlFnError,
    QueryBuilderErrorGroup,
    ToSqlFnFailure,
)


def analyze_query(query: Template) -> AnalyzedQuery:
    "Anlyze a template."
    return _QueryAnalyzer().analyze(query)


def _build_query(analyzed: AnalyzedQuery, conv_reg: ToSqlFnRegistry) -> BuiltQuery:
    "Build an analyzed query using the given converter registry."
    errors: list[ParamMissingToSqlFnError | ToSqlFnFailure] = []
    converted_params: dict[str, ConvertedParam] = {}

    for param in analyzed.params.values():
        try:
            if param.converter_name:
                converter = conv_reg._get_named(param.converter_name)
            else:
                converter = conv_reg._get_typed(type(param.original_value))
        except (MissingTypedToSqlFnError, MissingNamedToSqlFnError) as e:
            missing = ParamMissingToSqlFnError(param, e)
            missing.__cause__ = e

            errors.append(missing)
            continue

        try:
            value = converter(param.original_value)
        except (MissingTypedToSqlFnError, MissingNamedToSqlFnError) as e:
            # type dispatch happens later
            # so we need to get it here.
            missing = ParamMissingToSqlFnError(param, e)
            missing.__cause__ = e

            errors.append(missing)
            continue
        except Exception as e:
            adapter_err = ToSqlFnFailure(param)
            adapter_err.__cause__ = e

            errors.append(adapter_err)
            continue

        converted_params[param.param_name] = ConvertedParam(
            param_name=param.param_name,
            original_param_expr=param.original_param_expr,
            converter_name=param.converter_name,
            original_value=param.original_value,
            converted_value=value,
        )

    if errors:
        raise QueryBuilderErrorGroup(
            "Errors during query building", errors, analyzed, converted_params
        )
    return BuiltQuery(analyzed.query_str, converted_params)


class _QueryAnalyzer:
    def __init__(self):
        self.query_str = ""
        self._generated_name_counter = 0
        self.params: dict[str, Param] = {}

    def _gen_name(self) -> str:
        name = f"_CQL_{self._generated_name_counter}"
        self._generated_name_counter += 1
        return name

    def _make_name(self, expr: str, conv: str) -> str:
        match (expr.isidentifier(), conv):
            case (True, ""):
                return expr
            case (True, conv) if conv.isidentifier():
                return f"_CQL_{expr}_{conv}"
            case _:
                return self._gen_name()

    # TODO: double-escape _CQL if they used that
    # TODO: two interps next to each other has to be a bug, right?
    # it won't look like that in python which makes this even more pernicious, hmm
    def _add_part(self, part: str | Interpolation):
        if isinstance(part, str):
            # I believe this is as well-optimized as "".join(list[str]) ?
            self.query_str += part
            return
        # TODO: nicer error
        assert part.conversion is None, "conversions not yet supported, do it manually"
        if part.format_spec == "DANGER":
            self.query_str += str(part.value)
            return

        name = self._make_name(part.expression, part.format_spec)
        self.query_str += f":{name}"
        if name not in self.params:
            self.params[name] = Param(
                param_name=name,
                original_param_expr=part.expression,
                converter_name=part.format_spec,
                original_value=part.value,
            )

    @classmethod
    def analyze(cls, template: Template) -> AnalyzedQuery:
        self = cls()
        for part in template:
            self._add_part(part)
        return AnalyzedQuery(self.query_str, self.params)


__all__ = [
    "analyze_query",
    "AnalyzedQuery",
    "Param",
    "ConvertedParam",
    "BuiltQuery",
    "ParamMissingToSqlFnError",
    "ToSqlFnFailure",
    "QueryBuilderErrorGroup",
]



from contextlib import closing
from functools import wraps
from pathlib import Path
import sqlite3 as sql3
from string.templatelib import Template
from typing import Any, Literal, overload
import typing
from urllib.parse import parse_qs, urlsplit, urlunsplit
from weakref import WeakValueDictionary

import cattrs

from coolqlite._results import QueryResults, Row
from coolqlite._util import UNDEFINED, _Undef, _last
from coolqlite.from_sql import configure_converter
from coolqlite.to_sql import ToSqlFnRegistry
from coolqlite._common import (
    _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK,
    RowFactory,
    SqliteType,
)
from coolqlite.experimental.transactions import AdvancedSavepoint
from coolqlite.query import BuiltQuery, analyze_query, _build_query
from coolqlite.query._common import AnalyzedQuery
from coolqlite.transactions import (
    _SavepointStatus,
    IncorrectTransactionControlError,
    Savepoint,
)


def _conv_factory[T](converter: cattrs.BaseConverter, target: type[T]) -> RowFactory[T]:
    hook = converter.get_structure_hook(target)

    @wraps(hook)
    def hook_as_row_factory(cursor: sql3.Cursor, data: tuple[Any, ...]) -> T:
        row = Row(cursor, data)
        return hook(row, target)

    return hook_as_row_factory


def _column_conv_factory[T](
    converter: cattrs.BaseConverter, target: type[T]
) -> RowFactory[T]:
    hook = converter.get_structure_hook(target)

    @wraps(hook)
    def hook_as_column_row_factory(_cursor: sql3.Cursor, data: tuple[Any, ...]) -> T:
        return hook(data[0], target)

    return hook_as_column_row_factory


class Connection[RowType = Row]:
    """A connection to a sqlite database.

    Connect with :func:`connect`, not the constructor.

    It is generic over the RowType configured with :func:`connect`,
    which defaults to :class:`Row`.

    .. TODO:: More comprehensive Connection docs?
       Or should that be at the module level / in prose?
    """

    def __init__(
        self,
        private,
        connec: sql3.Connection,
        to_sql_fns: ToSqlFnRegistry,
        converter: cattrs.BaseConverter,
    ):
        assert private is _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK
        self._connec = connec
        self._to_sql_fns = to_sql_fns
        # self._savepoint_stack: list[str] = []
        self._savepoints: WeakValueDictionary[str, Savepoint] = WeakValueDictionary()
        self._conv = converter

    # TODO: actually, no row factories yet.
    # Let someone ask for it.

    # If type is specified, that's what you get.
    @overload
    def query[T](
        self, query: Template | AnalyzedQuery | BuiltQuery, type_: type[T], /
    ) -> QueryResults[T]: ...
    @overload
    def query(
        self, query: Template | AnalyzedQuery | BuiltQuery, /
    ) -> QueryResults[RowType]: ...

    def query[T](
        self,
        query: Template | AnalyzedQuery | BuiltQuery,
        type_: type[T] | Literal[_Undef.UNDEFINED] = _Undef.UNDEFINED,
        /,
    ) -> QueryResults[Any]:
        """Run the given query."""

        row = None if type_ is _Undef.UNDEFINED else _conv_factory(self._conv, type_)

        return self._inner_query(query, row)

    # TODO: can we just expose this instead???
    # so if you wanna customize a row factory you can,
    # but in its own method.
    # TODO: _maybe_ we can get typing guarantees here.
    def _inner_query(
        self,
        query: Template | AnalyzedQuery | BuiltQuery,
        row: RowFactory | None = None,
    ) -> QueryResults[Any]:
        # note: row = None is not the same as in sqlite3,
        # it means we ignore it instead of making it return a tuple.

        qb = query if isinstance(query, BuiltQuery) else self.build(query)

        cursor = self._connec.cursor()

        if row is not None:
            # pyright is pywrong: it thinks the second arg to a row factory is always a Row
            cursor.row_factory = row  # pyright: ignore[reportAttributeAccessIssue]

        cursor.execute(qb.query_str, qb._converted_dict())

        return QueryResults(_DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK, cursor)

    @overload
    def colquery(
        self, query: Template | AnalyzedQuery | BuiltQuery, /
    ) -> QueryResults[SqliteType]: ...
    @overload
    def colquery[T](
        self,
        query: Template | AnalyzedQuery | BuiltQuery,
        type_: type[T],
        /,
    ) -> QueryResults[T]: ...

    def colquery(
        self,
        query: Template | AnalyzedQuery | BuiltQuery,
        type_: Any = UNDEFINED,
        /,
    ) -> QueryResults[Any]:
        match type_:
            case _Undef.UNDEFINED:
                target_type = SqliteType
            case other:
                target_type = other
        # for some reason, pyright doesn't like the below
        # target_type = SqliteType if type_ is UNDEFINED else type_
        # TODO: shared "inner query" function so we don't have to double-check
        return self._inner_query(
            query, row=_column_conv_factory(self._conv, target_type)
        )

    # TODO: return rowcount? Maybe lastrowid or both?
    def run(self, query: Template | AnalyzedQuery | BuiltQuery):
        """
        Run a sql statement without looking at the results.
        """
        qb = self.build(query) if not isinstance(query, BuiltQuery) else query
        with closing(self._connec.cursor()) as cursor:
            cursor.row_factory = None
            cursor.execute(
                qb.query_str,
                {key: value.converted_value for key, value in qb.params.items()},
            )

    @property
    def user_version(self) -> int:
        """Access to the ``user_version`` sqlite pragma.

        You can use this to manage migrations in a simple fashion:

        .. code-block:: python

           MIGRATIONS = [file.read_text() for file in MIGRATIONS_DIR.iterdir()]
           with db:
               for migration in MIGRATIONS[db.user_version:]:
                   db.run_script(migration)

               db.user_version = len(MIGRATIONS)

        """
        with closing(self._connec.execute("PRAGMA user_version")) as q:
            return int(q.fetchone()[0])

    @user_version.setter
    def user_version(self, version: int):
        # we don't allow subclasses either. could be str() shenanigans.
        if type(version) is not int:
            raise TypeError("version must be an int")

        # TODO: this wouldn't work with autocommit=False.
        # The _config_ version would.
        # But we dont'support autocommit=False anyway?
        self._connec.execute(f"PRAGMA user_version = {version}").close()

    def close(self):
        """Close the connection to the database.
        This will roll back any currently active transaction.

        Note that using the Connection as a context manager *will not*
        close it at the end, since context management is used
        for transactions.

        Like the stdlib suggests, you can instead use :func:`contextlib.closing` :

        .. code-block:: python

           from contextlib import closing
           from coolqlite import connect

           with closing(connect(":memory:")) as db:
               with db: # transaction!
                   ... # do stuff here

        """
        self._connec.close()

    # TODO: we should force _all_ to be templates for safety,
    # even this. Just disallow any actual substitutions.
    # TODO: when do you want the cursor?
    def RUN_SCRIPT_DONT_USE_TEMPLATES(self, script: str):
        """Execute multiple statements.
        Does not give access to the cursor.

        .. warning:: This may be removed or reworked in the future;
           I'm unsure about its design.
        """
        self._connec.executescript(script).close()

    def build(self, query: Template | AnalyzedQuery) -> BuiltQuery:
        """Build the query, converting its values.

        See :mod:`coolqlite.query` for terminology,
        and :func:`coolqlite.query.build_query` for details.
        """
        analyzed = query if isinstance(query, AnalyzedQuery) else analyze_query(query)
        return _build_query(analyzed, self._to_sql_fns)

    def with_savepoint(self) -> Savepoint:
        """Create a new savepoint and a context manager for it.

        See the :mod:`coolqlite.transactions` module docs
        and :class:`~coolqlite.transactions.Savepoint` for details.
        """
        savepoint_number = len(self._savepoints)
        savepoint_id = f"_CQL_save_{savepoint_number}"
        self._connec.execute(f"savepoint {savepoint_id}")
        latest_savepoint = _last(self._savepoints.values(), None)
        self._savepoints[savepoint_id] = savepoint = Savepoint(
            _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK,
            savepoint_id,
            self,
            latest_savepoint,
        )
        return savepoint

    def with_EXPERIMENTAL_SAVEPOINT(self) -> AdvancedSavepoint:
        savepoint_number = len(self._savepoints)
        savepoint_id = f"_CQL_save_{savepoint_number}"
        self._connec.execute(f"savepoint {savepoint_id}")
        latest_savepoint = _last(self._savepoints.values(), None)
        self._savepoints[savepoint_id] = savepoint = AdvancedSavepoint(  # type: ignore
            _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK,
            savepoint_id,
            self,
            # TODO: intermingling regular and advanced?
            latest_savepoint,  # type: ignore
        )
        return savepoint

    def _invalidate_all_savepoints_due_to_on_conflict_rollback(self):
        if self._connec.in_transaction:
            raise IncorrectTransactionControlError(
                "Tried to invalidate all savepoints due to conflict "
                "but we're still in a transaction"
            )
        for savepoint in self._savepoints.values():
            savepoint._status = _SavepointStatus.DEAD

    def _release(self, savepoint_id: str):
        self._connec.execute(f"release {savepoint_id}").close()
        self._savepoints.pop(savepoint_id)

    def _rollback_to_before(self, savepoint_id: str):
        with closing(self._connec.cursor()) as c:
            c.execute(f"rollback to {savepoint_id}")
            c.execute(f"release {savepoint_id}")
        self._savepoints.pop(savepoint_id)

    def _rollback_to_after(self, savepoint_id: str):
        latest_txn_id = _last(self._savepoints.keys())
        if latest_txn_id != savepoint_id:
            raise IncorrectTransactionControlError(
                f"Tried to rollback to after {savepoint_id} but the"
                f"innermost savepoint id was {latest_txn_id}"
            )
        self._connec.execute(f"rollback to {savepoint_id}").close()

    # TODO: hide until we can force a template string and insist on no params.
    # TODO: well, except for DANGER passthroughs.
    # def query_posn(
    #     self,
    #     query: str,
    #     # technically not sequence but I can't get ty to behave
    #     params: Sequence = tuple(),
    # ) -> Cursor[RowType]:
    #     """
    #     Query using positional params. Does not do
    #     """
    #     return Cursor(self._connec.execute(query, params))

    # TODO: need to figure out conversion strategy
    # def query_many(
    #     self, query: str, params: Iterable[Mapping[str, Any]] | Iterable[Sequence]
    # ):
    #     return Cursor(self._connec.executemany(query, params))


@overload
def connect[RowType = Row](
    database: str,
    *,
    uri: bool = False,
    readonly: Literal[True] | None = None,
    row_factory: RowFactory[RowType] = Row,
    to_sql_fns: ToSqlFnRegistry | None = None,
    converter: cattrs.BaseConverter | None = None,
) -> Connection[RowType]: ...
@overload
def connect[RowType = Row](
    database: Path,
    *,
    readonly: Literal[True] | None = None,
    row_factory: RowFactory[RowType] = Row,
    to_sql_fns: ToSqlFnRegistry | None = None,
    converter: cattrs.BaseConverter | None = None,
) -> Connection[RowType]: ...


def connect[RowType = Row](
    database: str | Path,
    *,
    uri: bool = False,
    readonly: Literal[True] | None = None,
    row_factory: RowFactory[RowType] = Row,
    to_sql_fns: ToSqlFnRegistry | None = None,
    converter: cattrs.BaseConverter | None = None,
) -> Connection[RowType]:
    """
    Connect to a sqlite database.

    Works similarly to :func:`sqlite3.connect`,
    but has fewer flags exposed for the sake of good-practices defaults.

    See package docs for details.

    :param database: The path to the sqlite db file,
                     or you can use ``":memory:"`` for an in-memory db.
    :type database: str | Path
    :param row_factory: A custom :type:`RowFactory`, defaults to :class:`Row`.
    :type row_factory: RowFactory
    :param to_sql_fns: The registry of sql serialization functions.
                       If unspecified / None, uses the defaults for
                       :class:`~coolqlite.to_sql.ToSqlConfig` .
    :type to_sql_fns: ToSqlFnRegistry | None
    :return: The connection to the database,
             with the correct Row type as per ``row_factory`` .
    :rtype: Connection
    """
    return Connection(
        _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK,
        _sql3_connect(database, uri=uri, readonly=readonly, row_factory=row_factory),
        to_sql_fns=ToSqlFnRegistry() if to_sql_fns is None else to_sql_fns,
        converter=converter if converter is not None else configure_converter(),
    )


def _sql3_connect(
    database: str | Path,
    *,
    uri: bool,
    readonly: Literal[True] | None,
    row_factory: RowFactory,
) -> sql3.Connection:
    """Set up a connection with autocommit=True, foreign keys on,
    and the row factory set up.
    """
    match (database, uri, readonly):
        case (Path(), True, _):
            raise TypeError("Cannot specify uri=True for a Path")
        case (Path(), False, ro):
            uri_str = f"file:{database.as_posix()}" + "?mode=ro" if ro else ""
        case (str(), False, ro):
            uri_str = f"file:{database}" + "?mode=ro" if ro else ""
        case (str(), True, True):
            parsed = urlsplit(database)
            qs = parse_qs(parsed.query)
            if "mode" in qs:
                raise ValueError(
                    "Cannot specify readonly=True when mode is already in the URI"
                )
            qs["mode"] = ["ro"]
            uri_str = urlunsplit(parsed._replace(query=qs))
        case (str(), True, None):
            uri_str = database
        case unreachable:
            typing.assert_never(unreachable)
    connection = sql3.connect(uri_str, uri=True, autocommit=True)
    connection.execute("PRAGMA journal_mode=WAL").close()
    connection.execute("PRAGMA foreign_keys = ON").close()
    connection.row_factory = row_factory  # pyright: ignore[reportAttributeAccessIssue]
    return connection



from enum import Enum
from functools import wraps
from collections.abc import Callable
import inspect
from typing import Iterable, overload


class _Undef(Enum):
    UNDEFINED = "undefined"


UNDEFINED = _Undef.UNDEFINED


def renamed_fn(name: str):
    "Rename a function's __name__."

    def _curried(fn):
        @wraps(fn)
        def _wrapped(*args, **kwargs):
            return fn(*args, **kwargs)

        _wrapped.__name__ = name
        return _wrapped

    return _curried


# TODO: should this even be exposed?
# TODO: should this work for functions of any arg length, and specify
# it's only for the first one?
def type_checked[A, R](f: Callable[[A], R]) -> Callable[[A], R]:
    # TODO: handle this error for unannotated fn
    anno = next(iter(inspect.signature(f).parameters.values())).annotation

    @wraps(f)
    def _checked(arg: A) -> R:
        if not isinstance(arg, anno):
            raise TypeError(f"argument was not of type {anno}")
        return f(arg)

    return _checked


def _identity_fn[T](x: T) -> T:
    return x


@overload
def _last[T](ible: Iterable[T], /) -> T: ...
@overload
def _last[T, O = None](ible: Iterable[T], default: O = None, /) -> T | O: ...
def _last(ible: Iterable, default=_Undef.UNDEFINED, /):
    iter_ = iter(ible)

    try:
        item = next(iter_)
    except StopIteration:
        if default is not _Undef.UNDEFINED:
            return default
        raise

    for item in iter_:
        continue

    return item



from collections.abc import Callable, Generator, Iterable, Iterator
from contextlib import closing
from enum import Enum
from functools import wraps
from typing import Any, Protocol, overload, runtime_checkable
import sqlite3 as sql3

from coolqlite._common import _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK, CoolqliteError
from coolqlite._util import _identity_fn


class _RowCount(Enum):
    "The number of rows expected or received."

    Zero = "zero"
    One = "one"
    Any = "any"
    OneOrNone = "one or none"
    MoreThanOne = "more than one"

    def __str__(self):
        return self.value


class Row(sql3.Row):
    """
    Extends a :class:`sqlite3.Row` with the following niceities:

    - Easily allows accessing through a namespace
    - Requesting a missing column by name raises a :exc:`KeyError` instead
      of an :exc:`IndexError`
    - The ``repr`` is more readable
    """

    # if TYPE_CHECKING:

    #     def __new__(cls, cursor: sql3.Cursor, data: tuple[Any, ...]) -> Self: ...
    #     def __init__(self, cursor: sql3.Cursor, data: tuple[Any, ...]): ...

    @overload
    def __getitem__(self, key: str | int) -> Any: ...
    @overload
    def __getitem__(self, key: slice[Any, Any, Any]) -> tuple[Any, ...]: ...
    def __getitem__(self, key: str | int | slice[Any, Any, Any]):
        try:
            return super().__getitem__(key)
        except IndexError:
            if isinstance(key, str):
                raise KeyError(key) from None
            else:
                raise

    # TODO: they can do this themselves!
    # def namespace(self) -> types.SimpleNamespace:
    #     "Access columns as if they were attributes."
    #     return types.SimpleNamespace(self)

    def __repr__(self):  # pragma: no cover
        keys = self.keys()
        if not keys:
            return "Row()"
        key_exprs = [f"{key}={self[key]!r}" for key in keys]
        return f"<row with columns {', '.join(key_exprs)}>"


# @overload
# def mapped_row[T](mapper: Callable[[Row], T]) -> RowFactory[T]: ...
# @overload
# def mapped_row[T, R](
#     mapper: Callable[[R], T], row_factory: RowFactory[R]
# ) -> RowFactory[T]: ...


# # TODO: none should be tuple, no?
# def mapped_row(mapper: Callable, row_factory: RowFactory | None = None) -> RowFactory:
#     """
#     Compose a conversion over a :type:`RowFactory`, making it a :type:`RowFactory` itself.

#     Defaults to :class:`Row` if the wrapped ``row_factory`` is not specified.

#     For something more substantial, see :mod:`coolqlite.cattrs_` .
#     """
#     fac = row_factory if row_factory is not None else Row
#     # TODO: wrapper to update __name__ or whatsits?
#     return lambda cursor, data: mapper(fac(cursor, data))


# TODO: attach rows?
class ResultCountError(CoolqliteError):
    "A different number of rows than expected were returned."

    def __init__(self, *, expected: _RowCount, actual: _RowCount):
        super().__init__(f"Expected {expected} rows but got {actual}")
        self.expected = expected
        self.actual = actual


@runtime_checkable
class CloseableIterator[T](Iterator[T], Protocol):
    def close(self): ...


class _CursorIterator[RowType](Iterator[RowType]):
    def __init__(self, sql_cursor: sql3.Cursor):
        self._inner = sql_cursor

    def __next__(self) -> RowType:
        next_value = self._inner.fetchone()

        if next_value is None:
            # TODO: is this done for us?
            self._inner.close()
            raise StopIteration

        return next_value

    def close(self):
        self._inner.close()


def first_col[T](x: tuple[T, ...]) -> T:
    """Get the first item in a tuple.

    This is only here to help with type checkers complaining
    about being passed a union type for
    :meth:`~coolqlite.Connection.colquery` :

    You use :meth:`~coolqlite.Connection.query` instead,
    wrap the union in a tuple,
    and then give this to :meth:`Results.map` .

    .. code-block:: python

        result = db.query(t"select 1", tuple[int | None]).map(first_col)
        typing.assert_type(result, QueryResults[int | None])
    """
    return x[0]


# TODO: frozen?
class Results[RowType](Iterator[RowType]):
    """An iterator of rows that offers easy ways to
    validate the number of returned rows,
    and is parameterized over the type of the row
    (from the row factory).
    """

    @overload
    def __init__(
        self,
        inner: Iterator[RowType] | Iterable[RowType],
    ): ...
    @overload
    def __init__[Inner](
        self,
        inner: Iterator[Inner] | Iterable[Inner],
        mapping_fn: Callable[[Inner], RowType],
    ): ...

    def __init__(
        self,
        inner: Iterator | Iterable,
        mapping_fn: Callable = _identity_fn,
    ):
        self._close: Callable[[], None]
        match inner:
            case Generator():
                self._inner = inner
                self._close = inner.close
            case Iterator():
                self._inner = inner
                if hasattr(inner, "close") and isinstance(
                    (close := getattr(inner, "close")), Callable
                ):
                    self._close = close
                else:
                    self._close = lambda: None
            case Iterable():
                inner = iter(inner)
                # oh my god this feels dirty
                self.__init__(inner, mapping_fn)
                # close is not assigned because this'll get to
                # the Iterator case and do it there.

        self._closed = False
        self._mapping_fn = mapping_fn

    def __enter__(self):
        return self

    def __exit__(self, *_exc_deets):
        self.close()
        return False

    # TODO: what if it's wrapping a generator,
    # it exhausts itself, and then we call close on it?
    # we have to handle that case, right?

    def __next__(self) -> RowType:
        return self._mapping_fn(next(self._inner))

    def next(self) -> RowType | None:
        "Fetch the next row, or None if no rows remain."
        return next(self, None)

    def one(self) -> RowType:
        """Return exactly one row,
        raising a :exc:`QueryResultCountError`
        if there's zero or more than one.

        If you want *at least one*,
        then use ``next(cursor.any())``.
        """
        with closing(self):
            row = self.next()
            if row is None:
                raise ResultCountError(expected=_RowCount.One, actual=_RowCount.Zero)
            _next = self.next()
            if _next is not None:
                raise ResultCountError(
                    expected=_RowCount.One, actual=_RowCount.MoreThanOne
                )
            return row

    def one_or_none(self) -> RowType | None:
        """Return exactly one or zero rows,
        raising a :exc:`QueryResultCountError`
        if there's two or more.

        Useful for checking for the presence or absence of something.
        """
        with closing(self):
            row = self.next()
            if row is None:
                return None
            _next = self.next()
            if _next is not None:
                raise ResultCountError(
                    expected=_RowCount.OneOrNone, actual=_RowCount.MoreThanOne
                )
            return row

    def at_least_one(self) -> Iterator[RowType]:
        """Get an iterator that is guaranteed to yield at least one row.

        Raises a :exc:`QueryResultCountError` if there are none.
        """
        with closing(self):
            row = self.next()
            if row is None:
                self.close()
                raise ResultCountError(expected=_RowCount.Any, actual=_RowCount.Zero)
            yield row
            yield from self  # closes for us but doubled, uh?

    # IF SUBCLASSING...
    # what's the _right_ way to do this?
    # I guess it inherently needs to be customized, since
    # type(self) could always have a different signature?
    def map[R](self, map: Callable[[RowType], R], /) -> Results[R]:
        """Return a new :class:`Results` that applies the given function
        to every row.
        """
        return Results(self, map)

    # TODO: we can remove this, right?
    # just do a `select count(*)`?

    # def zero(self) -> None:
    #     "If there are any rows at all, raise a :exc:`QueryResultCountError` ."
    #     with closing(self):
    #         if self.next() is not None:
    #             raise QueryResultCountError(
    #                 expected=_RowCount.Zero, actual=_RowCount.Any
    #             )

    def close(self):
        """If the underlying iterator is closeable
        (has a ``close`` method), and hasn't been closed yet,
        close it.

        See :meth:`sqlite3.Cursor.close` if this was based off
        of a sqlite cursor..
        """
        # TODO: do we need to track? or is closing again a no-op?
        if not self._closed:
            self._close()
            self._closed = True


class QueryResults[RowType = Row](Results[RowType]):
    """An extension of :class:`Results` that allows
    access to cursor-specific information,
    like row count and ``lastrowid``.

    Note that the same semantics as in the ``sqlite3`` module apply!
    """

    def __init__(
        self,
        private,
        cursor: sql3.Cursor,
        mapping_fn: Callable[..., RowType] = _identity_fn,
    ):
        assert private is _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK
        super().__init__(_CursorIterator(cursor), mapping_fn)
        self._cursor = cursor

    @property
    def rowcount(self):
        "The underlying cursor's ``rowcount``."
        return self._cursor.rowcount

    @property
    def lastrowid(self):
        "The underlying cursor's ``lastrowid``."
        return self._cursor.lastrowid

    def map[R](self, map: Callable[[RowType], R], /) -> QueryResults[R]:
        """Return a new :class:`QueryResults` that applies the given function
        to every row.
        """
        qr = QueryResults(
            _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK,
            self._cursor,
            lambda x: map(self._mapping_fn(x)),
        )
        # TODO: we'd need this on the ctor if we make it frozen eventually
        return qr


def as_results[T, **P](fn: Callable[P, Iterable[T]]) -> Callable[P, Results[T]]:
    """Wrap a generator function so it returns a :class:`Results` instead.

    Useful for when you want to do some in-python processing over query results,
    but still want the caller to have control over how many rows they care about.

    .. code-block:: python

        from coolqlite import as_results, Connection

        @as_results
        def advanced_query(db: Connection):
            # imagine this is something more complex
            # than could easily be done in SQL.
            for name in db.colquery(t"select name from Table", str):
                if name.startswith("Alex"):
                    yield name[::-1]
                else:
                    yield name

        only_one = advanced_query(db).one()
    """

    @wraps(fn)
    def wrapped(*args: P.args, **kwargs: P.kwargs):
        return Results(fn(*args, **kwargs))

    return wrapped



"""
Helpers to convert DateTimes to sqlite types.

.. name table:

==============================  =======================  =======================================
Function Name                   Default Named Converter  Accepts
==============================  =======================  =======================================
:meth:`aware_as_zulu_string`    ``z``                    AwareDateTime, AwareTime if already UTC
:meth:`aware_as_keeptz_string`  ``keeptz``               AwareDateTime, AwareTime
:meth:`naive_as_string`         ``naive``                NaiveDateTime, NaiveTime, Date
:meth:`any_temporal_as_iso`     ``ANY_temporal``         Any of the above
==============================  =======================  =======================================
"""

from datetime import timedelta, timezone
from datetype import (
    NaiveDateTime,
    AwareDateTime,
    Date,
    NaiveTime,
    AwareTime,
)

from coolqlite._common import ToSqlTypeError, ToSqlValueError


def any_temporal_as_iso(
    temporal: NaiveDateTime | AwareDateTime | Date | NaiveTime | AwareTime, /
) -> str:
    # pyright no likey datetype stuff for some reason
    match temporal:
        case AwareDateTime() | NaiveDateTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat(sep=" ")
        case Date() | AwareTime() | NaiveTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat()
        case _:
            raise ToSqlTypeError(f"{type(temporal)} is not a temporal type")


def aware_as_zulu_string(temporal: AwareDateTime | AwareTime, /) -> str:
    """Serialize an :class:`datetype.AwareDateTime`
    by adjusting it to UTC, then serializing
    as an iso-format string with a ``Z`` at the end.
    """
    match temporal:
        case AwareDateTime():  # pyright: ignore[reportGeneralTypeIssues]
            return (
                temporal.astimezone(timezone.utc).replace(tzinfo=None).isoformat(" ")
                + "Z"
            )
        case AwareTime() if temporal.utcoffset() == timedelta():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.replace(tzinfo=None).isoformat() + "Z"
        case AwareTime():  # pyright: ignore[reportGeneralTypeIssues]
            raise ToSqlValueError(
                f"Aware times can't be converted to UTC cleanly: {temporal}"
            )
        case NaiveDateTime() | NaiveTime() | Date():  # pyright: ignore[reportGeneralTypeIssues]
            raise ToSqlTypeError(f"Expected aware temporal type, got naive: {temporal}")
        case _:
            raise ToSqlTypeError(f"{type(temporal)} is not a temporal type")


# TODO: fix docs


def aware_as_keeptz_string(temporal: AwareDateTime | AwareTime, /) -> str:
    """Serialize an :class:`datetype.AwareDateTime`
    as an iso-format string.
    """
    match temporal:
        case AwareDateTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat(" ")
        case AwareTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat()
        case NaiveDateTime() | NaiveTime() | Date():  # pyright: ignore[reportGeneralTypeIssues]
            raise ToSqlTypeError(f"Expected aware temporal type, got naive: {temporal}")
        case _:
            raise ToSqlTypeError(f"{type(temporal)} is not a temporal type")


def naive_as_string(temporal: NaiveDateTime | NaiveTime | Date, /) -> str:
    """Serialize a :class:`datetype.NaiveDateTime`
    as an iso-format string.
    """
    match temporal:
        case Date() | NaiveTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat()
        case NaiveDateTime():  # pyright: ignore[reportGeneralTypeIssues]
            return temporal.isoformat(" ")
        case AwareDateTime() | AwareTime():  # pyright: ignore[reportGeneralTypeIssues]
            raise ToSqlTypeError(f"Expected naive temporal type, got aware: {temporal}")
        case _:
            raise ToSqlTypeError(f"{type(temporal)} is not a temporal type")



"Vanilla common values and exceptions."

from collections.abc import Callable
import sqlite3 as sql3

SqliteType = None | int | float | str | bytes
"A type that sqlite natively understands."

# TODO: why no doc?
type RowFactory[Result] = Callable[[sql3.Cursor, tuple[SqliteType, ...]], Result]
"""A `RowFactory <https://docs.python.org/3/library/sqlite3.html#how-to-create-and-use-row-factories>`
"""


class CoolqliteError(Exception):
    "A base exception for all standard coolqlite errors."

    pass


class CoolqliteWarning(CoolqliteError, UserWarning):
    pass


class InternalCoolqliteError(CoolqliteError):
    def __new__(cls, msg: str = ""):  # pragma: no cover
        msg = f"YOU SHOULD NEVER SEE THIS I'M SORRY\nPlease file a ticket.\n{msg}"
        return super().__new__(cls, msg)


class CoolqliteExceptionGroup[E: Exception = Exception](
    ExceptionGroup[E], CoolqliteError
):
    "A base exceptiongroup for all standard coolqlite errors."

    pass


# TODO: a proper unified error handling philosophy,
# including thinking about how it works with cattrs.


class MissingNamedToSqlFnError(CoolqliteError):
    "A named adapter was missing."

    def __init__(self, name: str):
        super().__init__(f"Missing adapter named {name}")
        self.add_note("Did you forget to register it, or disable a default?")


class MissingTypedToSqlFnError(CoolqliteError):
    "A value was given of a type the converter didn't know how to handle."

    pass


# TODO: base type and make its subclasses final?
# is that possible like in kotlin?
type MissingToSqlFnError = MissingTypedToSqlFnError | MissingNamedToSqlFnError
"Some sort of ToSql converter was missing."


class ToSqlFnSetupError(CoolqliteError):
    "An error occurred while setting up a tosql function."

    pass


class CoolqliteTypeError(CoolqliteError, TypeError):
    "A type error."

    pass


class ToSqlTypeError(CoolqliteTypeError):
    "A type error related to converting to sql."

    pass


class ToSqlValueError(CoolqliteError, ValueError):
    "A value error related to converting to sql."

    pass


class FromSqlTypeError(CoolqliteTypeError):
    "A type error related to converting from sql."

    pass


_DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK = object()



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any
from coolqlite._common import (
    CoolqliteError,
    CoolqliteExceptionGroup,
    MissingToSqlFnError,
)


@dataclass(kw_only=True)
class Param:
    "An analyzed input parameter."

    param_name: str
    """The name assigned to the parameter.

    May be mangled if a filter was applied
    or it wasn't a valid identifier.
    """

    original_param_expr: str
    "The original expression used in the t-string."

    converter_name: str
    """The name of the converter to use.

    (Has not been validated to be present,
    that happens during building.)
    """
    original_value: Any
    "The value of the original expression in the t-string."


@dataclass(kw_only=True)
class ConvertedParam(Param):
    "A param that has been converted."

    converted_value: Any
    "The converted value."


@dataclass(eq=False)
class AnalyzedQuery:
    """A template that has been analyzed
    into the query string and
    (possibly name-mangled) parameters,
    ready to be converted.
    """

    query_str: str
    params: dict[str, Param]
    """Maps (possibly mangled) param name to the param."""

    def __str__(self):  # pragma: no cover
        return self.query_str


# while AnalyzedQuery and BuiltQuery could just be the
# same class but generic,
# I want to reduce the chance of them getting mixed up.
@dataclass(eq=False)
class BuiltQuery:
    """A query that has been "built"
    from an :class:`AnalyzedQuery`,
    with its parameters converted.
    """

    query_str: str
    params: dict[str, ConvertedParam]
    """Maps (possibly mangled) param name to the param."""

    def _converted_dict(self) -> dict[str, Any]:
        return {key: value.converted_value for key, value in self.params.items()}

    def __str__(self):  # pragma: no cover
        return self.query_str


@dataclass
class ParamMissingToSqlFnError(CoolqliteError):
    """A converter fn was missing.
    See error message for details.
    """

    param: Param
    "The parameter whose converter couldn't be found."

    cause: MissingToSqlFnError
    """
    This *should* be the same as ``__cause__`` but
    this helps guarantee it's there.
    """


@dataclass
class ToSqlFnFailure(CoolqliteError):
    """A converter itself threw an exception."""

    param: Param
    "The parameter being processed when this error occurred."


class QueryBuilderErrorGroup(
    CoolqliteExceptionGroup[ParamMissingToSqlFnError | ToSqlFnFailure]
):
    """Errors that occurred during query building."""

    def __new__(cls, message, exceptions, *_args, **kwargs):
        return super().__new__(cls, message, exceptions)

    def __init__(
        self,
        message: str,
        exceptions: Sequence[ParamMissingToSqlFnError | ToSqlFnFailure],
        input_: AnalyzedQuery,
        converted: dict[str, ConvertedParam],
    ):
        super().__init__(message, exceptions)
        self.input_ = input_
        self.converted = converted



# from collections.abc import Callable

# from coolqlite._common import SqliteType


# type UserFunction = Callable[..., SqliteType]


# def foo(x: SqliteType) -> int:
#     if not isinstance(x, int):
#         raise TypeError("oops")
#     return x


# def takes_user_fn(fn: UserFunction):
#     pass


# takes_user_fn(foo)

# import sqlite3 as sql3

# db = sql3.connect(":memory:")
# db.create_function()

# TODO: FOR USER FUNCTIONS
# need to enable passing exceptions upwards, apparently?
# why isn't that the default?
# there was some discussion about it.
# https://discuss.python.org/t/enable-unraisable-exceptions-by-default-for-sqlite3-callbacks/43808/20

# TODO: preprocess functions to validate the signature makes sense.
# maybe we need a connection preparer???



# TODO: MODULE DOCS FR FR

# TODO: should we just use cattrs?
# would kinda expose implementation deets...
# TODO: allow defining a fallback?
# TODO: allow access to conversion flag?
# TODO: flag for disallowing all defaults?
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
import functools
from typing import Any, overload
import warnings

from coolqlite._common import (
    CoolqliteWarning,
    InternalCoolqliteError,
    MissingNamedToSqlFnError,
    MissingTypedToSqlFnError,
    SqliteType,
    ToSqlFnSetupError,
)
from coolqlite._util import _identity_fn
from coolqlite.to_sql.datetime_ import (
    any_temporal_as_iso,
    aware_as_zulu_string,
    aware_as_keeptz_string,
    naive_as_string,
)

type ToSqliteConversionFn[Src = Any] = Callable[[Src], SqliteType]

_KEYWORDS = {"DANGER"}


class UuidAs(Enum):
    """How do you want to serialize a UUID?

    Right now only strings are offered, this may be expanded.
    """

    STRING = "string"
    # TODO: idk right now
    # BYTES = "bytes"


@dataclass(kw_only=True, frozen=True)
class ToSqlConfig:
    """Configure the ToSqlFnRegistry."""

    configure_datetime_fns: bool = True
    "Add named functions to convert various datetime types to sql."


def _builtins_from_cfg(config: ToSqlConfig) -> _BuiltinFns:
    builtins = _BuiltinFns()
    if config.configure_datetime_fns:
        builtins["z"] = aware_as_zulu_string
        builtins["keeptz"] = aware_as_keeptz_string
        builtins["naive"] = naive_as_string
        builtins["ANY_temporal"] = any_temporal_as_iso
    return builtins


class ToSqlFnRegistry:
    """
    Dispatch to serializers based on name or type.

    Register type-based dispatch as you would a :func:`functools.singledispatch`:

    .. code-block:: python

       registry = ToSqlFnRegistry()
       @registry.register
       def _handle_int(x: int) -> str:
           return str(x)

    Or register by name:

    .. code-block:: python

       registry = ToSqlFnRegistry()

       @registry.add_named("reversed-int")
       def _reversed_int(x: int) -> str:
           return str(x).reverse()

       registry.add_named("upper", str.upper)

    """

    def __init__(self, *, config: ToSqlConfig = ToSqlConfig()):
        builtins = _builtins_from_cfg(config)
        self._named = _NamedFnRegistry(builtins)

        @functools.singledispatch
        def by_type(x: object) -> Any:
            raise MissingTypedToSqlFnError(
                f"No type-dispatched adapter was registered for {type(x)}"
            )

        self._by_type = by_type
        self.register = self._by_type.register

        # base types
        # WHAT DO YOU MEAN IT DOESN'T MATCH MYPY WHAT ARE YOU SMOKING
        by_type.register(SqliteType, _identity_fn)  # type: ignore[call-overload]

    # TODO: don't quite like having named and typed both get and run,
    # but it's helpful for figuring out where the error is from.
    # Something _feels_ off though.
    # TODO: especially because singledispatch doesn't let us know until it's too late!!!
    # cattrs defaults to identity for unstructure so I guess they don't handle it specially either.
    # structure, they know.
    # but wait, maybe that's configurable?

    def _get_named(self, name: str) -> ToSqliteConversionFn:
        if not name:
            return _identity_fn
        try:
            return self._named[name]
        except KeyError as e:
            raise MissingNamedToSqlFnError(name) from e

    def _run_named(self, name: str, arg: Any) -> SqliteType:
        return self._get_named(name)(arg)

    def _get_typed(self, type_: type) -> ToSqliteConversionFn:
        return self._by_type.dispatch(type_)

    def _run_typed(self, value: Any) -> Any:
        return self._by_type(value)

    # TODO: more specific error type?
    def _set_named[F: ToSqliteConversionFn](
        self, key: str, f: F, warning_stacklevel: int
    ) -> F:
        if not key:
            raise ToSqlFnSetupError("need an adapter name")
        if not key.isidentifier():
            err = ToSqlFnSetupError(f"must be a valid python identifier, {key} is not.")
            err.add_note("this may be relaxed in the future.")
            raise err
        if key in _KEYWORDS:
            raise ToSqlFnSetupError(f"{key} is a built-in keyword")
        self._named.set(key, f, warning_stacklevel)
        return f

    # TODO: test warning stack level is correct
    @overload
    def add_named[F: ToSqliteConversionFn](self, name: str) -> Callable[[F], F]: ...
    @overload
    def add_named(self, name: str, fn: ToSqliteConversionFn) -> None: ...
    def add_named[F: ToSqliteConversionFn](
        self, name: str, fn: F | None = None
    ) -> Callable[[F], F] | None:
        """
        Add a named converter. Usable as a decorator.
        """
        if fn is None:
            return lambda fn: self._set_named(name, fn, 4)
        else:
            self._set_named(name, fn, 4)
            return None


class _BuiltinFns(dict[str, ToSqliteConversionFn]):
    def __setitem__(self, key: str, value: ToSqliteConversionFn):
        if key in self:  # pragma: no cover
            raise InternalCoolqliteError(
                "Attempting to register a builtin with the same name."
            )
        super().__setitem__(key, value)


class _NamedFnRegistry:
    def __init__(self, builtins: _BuiltinFns):
        self.builtins = builtins
        self.user: dict[str, ToSqliteConversionFn] = {}

    # def __contains__(self, key: object) -> bool:
    #     return key in self.user or key in self.builtins

    def __getitem__(
        self, key: str
    ) -> Callable[[Any], int | float | str | bytes | None]:
        if key in self.user:
            return self.user[key]
        if key in self.builtins:
            return self.builtins[key]
        raise KeyError(key)

    def set(self, key: str, value: ToSqliteConversionFn, warning_stacklevel: int):
        if key in self.user:
            warnings.warn(
                f"You are shadowing an already-set user function {key}",
                CoolqliteWarning,
                stacklevel=warning_stacklevel,
            )
        elif key in self.builtins:
            warnings.warn(
                f"You are shadowing a built-in / configured function {key}",
                CoolqliteWarning,
                stacklevel=warning_stacklevel,
            )
        self.user[key] = value


__all__ = ["ToSqlConfig", "ToSqlFnRegistry", "UuidAs"]



"""
For advanced usage, let's make sure you understand savepoints first.

About Savepoints
================

You shouldn't just think of savepoints as "named, nested commits",
because the terminology is slightly different.
You should get comfortable with the
`official sqlite documentation for savepoints <https://sqlite.org/lang_savepoint.html>`__.

Essentially, savepoints are a stack of frames, each holding some amount of database changes.
The first created savepoint is the "outermost" savepoint,
the most recently created savepoint is the "innermost" savepoint,
and the ones in between are "outer" and "inner" relative to each other.

When you want to okay your changes, instead of "committing", you "release" that frame.
If there's an outer frame, that means the changes are now "merged" into that frame,
like removing a fence sectioning off a field.
If the frame you released is already the outermost, last one, then those changes
are written to the database.

With plain ``BEGIN..COMMIT`` transactions, we tend to think of rollbacks as
"cancelling". And indeed, a vanilla ``ROLLBACK`` on its own will cancel everything.
But when we use ``rollback to SAVEPOINT_NAME``, it's more like
"restarting" that savepoint. All changes within that frame are discarded,
but the frame remains active! For clarity, we refer to this as
"rolling back to *after*" the savepoint.

You can still have the "cancellation" behavior by immediately releasing
a just-rolled-back savepoint. We call this "rolling back *before*" the savepoint.

Now that we understand how savepoints work in sqlite,
let's see how they work in coolqlite.

Advanced Usage
==============

.. warning:: This part of the API is somewhat experimental.

We try to map an active savepoint to a context manager with-block 1:1.

This means that if you do anything that would make the savepoint
no longer active (releasing it, rolling back before it,
rolling back before or after an outer savepoint),
this should be viewed as an "early exit" from that with-block.

This is accomplished with a :exc:`TransactionControlException`.

.. admonition:: Isn't it bad to use exceptions for control flow?

    Because Exceptions aren't reflected in a function's signature
    in many languages, using them for control flow is considered
    *surprising* and *hidden* behavior.

    Here it's not as hidden and surprising, because this is only ever raised
    by methods you yourself called in ``ALL_CAPS``.
    Also, control flow only ever jumps to after the with-block holding
    the context manager you called that method on.

    Also, it is the only way to model a conditional early exit from a
    context manager-- that sounds a lot like what an exception does to me.

    Just think of it as a special ``return`` that is scoped to that
    with-block instead of a function.

.. warning:: This means it is **always incorrect** to catch a
    :exc:`TransactionControlException` without re-raising it.

    You may need to handle this specially if writing your own
    context manager or catching exceptions that happen within
    a :class:`Savepoint` context manager.

An innermost savepoint can do the following:

1. Roll back to *after* the savepoint was started,
   discarding all changes since then,
   but staying within the savepoint.

2. Roll back to *before* the savepoint was started,
   discarding all changes since then,
   and *exiting* / erasing that savepoint

3. Release the savepoint, committing changes to the database
   and *exiting* / erasing this savepoint.

Following the 1:1 rule, this means that 2 and 3 should exit
the with-block, but 1 should stay in it.

With this in mind, let's do a similar database setup.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: setup
    :end-before: # endregion setup


Scenario 1, rolling back to after, is straightforward:

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: rollback after
    :end-before: # endregion rollback after

You'll understand why the method is called that later.


Scenario 2, rolling back before, is a little more unusual.
Since it means exiting the savepoint, we exit the with-block
early with an exception. Because that happens, the method is
in ``ALL_CAPS`` to call out how it's different.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: rollback before
    :end-before: # endregion rollback before


Scenario 3, releasing (like committing), is similar.
Because it means you're "exiting" the savepoint,
we exit the with-block early again with an exception.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: early release
    :end-before: # endregion early release


Nested Savepoints
-----------------

A savepoint that isn't the innermost can do all of the above,
with one caveat. Let's go backwards.

Scenario 3, releasing, works similarly when called on an
outer savepoint.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: nested release
    :end-before: # endregion nested release

Scenario 2, rolling back before, also similar.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: nested rollback before
    :end-before: # endregion nested rollback before

Scenario 1, rolling back after, is where things get tricky.

If you're nested within an outer savepoint, what would it mean to
rollback after it? The outer savepoint keeps going, sure, but
that means your nested one should be erased. Sure enough, that means
the inner one needs to exit early with an exception.

Note that we're calling this method on the savepoint we want to
reset-after *to*, not the inner nested one!

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: nested release after
    :end-before: # endregion nested release after

Caveat: ON CONFLICT ROLLBACK
============================

There is one tricky case that we're still figuring out how to handle well.

We mentioned how an unqualified ``ROLLBACK`` will tear down all savepoints.
Well, that applies to rollbacks that happen from ``ON CONFLICT ROLLBACK``
triggers on unique columns, as well as ``INSERT OR ROLLBACK`` queries.

.. TODO:: verify the insert part. Is has to be the same, right?

Recall that in our example we made the name column
``UNIQUE ON CONFLICT ROLLBACK``.

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: good rollback from violation
    :end-before: # endregion good rollback from violation

This is handled correctly by the exception handler for the savepoint
context manager. In fact, only one of them needs to see it to properly
invalidate all of the others!

But, what if you somehow kept going in a savepoint with-block?

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: complication from rollback
    :end-before: # endregion complication from rollback

.. TODO:: Improve the error message on that.

An invalidated savepoint context manager knows to not touch anything
on exit. But we can't stop you from using the methods.
For this reason, we encourage you to let those errors bubble up
to the outermost savepoint.

We may offer a method to deliberately "jump" to the outermost transaction
when this happens, as we can more or less detect when this happens.
But I'm still thinking about the API for that.

But even worse is when the context manager doesn't see it at all:

.. literalinclude:: /../examples/savepoints_advanced.py
    :start-after: # region: oh god it's even worse
    :end-before: # endregion oh god it's even worse

.. TODO:: I think I can handle that better with similar detection techniques.
"""

import typing
import sqlite3 as sql3
from functools import wraps
from typing import TYPE_CHECKING, Any, Never, Self

from coolqlite._common import _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK, CoolqliteError
from coolqlite.transactions import _SavepointStatus, IncorrectTransactionControlError


if TYPE_CHECKING:
    from coolqlite._connection import Connection


class TransactionControlException(CoolqliteError):
    """An exception used when an operation on a savepoint
    should alter control flow.

    .. warning:: It is **always** incorrect to catch this
       without re-raising it.

       You may need to handle this specially if writing your own
       context manager or catching exceptions that happen within
       a :class:`Savepoint` context manager.

    .. admonition:: Isn't it bad to use exceptions for control flow?

       Because Exceptions aren't reflected in a function's signature
       in many languages, it's considered to be surprising hidden behavior.

       It's not that hidden or surprising, because this is only ever raised
       by methods you yourself called in ``ALL_CAPS``.

       Also, it is the only way to model a conditional early exit from a
       context manager-- that sounds a lot like what an exception does to me.

    .. warning:: To show that this is special, we may change this to inherit
        from :exc:`BaseException` in the future.
    """

    # Oh yeah, three admonitions. That's how you know it's a good idea.

    pass


class _RollbackToAfterNotInnermost(TransactionControlException):
    __match_args__ = ("id_",)

    def __init__(self, id: str):
        super().__init__()
        self.id_ = id

    def __repr__(self):  # pragma: no cover
        return f"<ROLLING BACK TO AFTER (not innermost): {self.id_}>"


class _RollbackToBefore(TransactionControlException):
    __match_args__ = ("id_",)

    def __init__(self, id: str):
        super().__init__()
        self.id_ = id

    def __repr__(self):  # pragma: no cover
        return f"<ROLLING BACK TO BEFORE: {self.id_}>"


class _Release(TransactionControlException):
    __match_args__ = ("id_",)

    def __init__(self, id: str):
        super().__init__()
        self.id_ = id

    def __repr__(self):  # pragma: no cover
        return f"<RELEASING: {self.id_}>"


def _must_be_active(fn):
    @wraps(fn)
    def _active_check(self: AdvancedSavepoint, *args, **kwargs):
        if self._status is not _SavepointStatus.ACTIVE:
            raise IncorrectTransactionControlError(
                "Attempted to call a savepoint method on a savepoint"
                f"that was not active, but rather was {self._status}"
            )
        return fn(self, *args, **kwargs)

    return _active_check


class AdvancedSavepoint:
    """A context manager representing a sqlite savepoint.

    Create with the :meth:`coolqlite.Connection.with_savepoint` method.

    .. warning:: *Only* use this as a context manager,
       the methods assume you are and things will break if you don't.

    .. warning:: Methods in ``ALL_CAPS`` use
        :exc:`TransactionControlException` to work and enforce correctness.

        Understand then when writing code within the ``with`` block,
        using other context managers that interact with exceptions,
        and when using ``try``/``except``/``finally``.

        In other words, the exceptions they throw are not *problems*,
        unless you mishandle them.

        The methods are properly type-annotated, so tools should show
        that code that follows is unreachable.

        See the :mod:`module docs <coolqlite.transactions>`
        for details.
    """

    def __init__(
        self,
        private,
        id: str,
        connec: Connection[Any],
        parent: AdvancedSavepoint | None,
    ):
        assert private is _DONT_INIT_YOURSELF_OR_THINGS_WILL_BREAK
        self._id = id
        self._status: _SavepointStatus = _SavepointStatus.JUST_BORN
        self._connec = connec
        self._parent = parent

    # note: exceptions raised in enter do not trigger exit
    # (what would they even call it on? not all managers return self)
    def __enter__(self) -> Self:
        match self._status:
            case _SavepointStatus.JUST_BORN:
                self._status = _SavepointStatus.ACTIVE
                return self
            case _SavepointStatus.ACTIVE:
                raise IncorrectTransactionControlError(
                    "tried to enter the same savepoint more than once"
                )
            case _SavepointStatus.DEAD:
                raise IncorrectTransactionControlError(
                    "tried to enter an already-used savepoint"
                )

    def __exit__(self, _exc_type, exc_val: BaseException | None, _tb) -> bool:
        # I guess we also might want an exception for
        # "I know we're invalidated just get me out of here".
        dead = self._status is _SavepointStatus.DEAD
        match (dead, exc_val):
            case (True, None):
                return True
            case (False, None):
                self._do_release()
                return True
            case (True, TransactionControlException()):
                raise IncorrectTransactionControlError(
                    "Somehow got a transaction control exception after being invalidated"
                )
            case (False, _Release(id_)):
                self._do_release()
                # if this was for us, then we're done.
                return id_ == self._id
            case (False, _RollbackToBefore(id_=self._id)):
                self._do_rollback_before()
                return True
            case (False, _RollbackToAfterNotInnermost(id_)) if (
                self._parent is not None and id_ == self._parent._id
            ):
                self._do_rollback_before()
                self._parent._do_rollback_after()
                return True
            case (False, _RollbackToAfterNotInnermost(id_=self._id)):
                raise IncorrectTransactionControlError(
                    "This method should only be called on non-innermost transactions"
                )
            case (True, _):
                # some other exception while we're dead
                return False
            case (False, sql3.IntegrityError()) if (
                not self._connec._connec.in_transaction
            ):
                # this was a ON CONFLICT ROLLBACK,
                # so all savepoints are now invalidated.
                self._connec._invalidate_all_savepoints_due_to_on_conflict_rollback()
                return False
            case (False, _):
                # either a third-party exception
                # or a TransactionControl that wasn't for us.
                # in either case, rollback-to-before
                # but keep it going.
                self._do_rollback_before()
                return False
            case unreachable:  # pragma: no cover
                # TODO: pyright correctly reduces tuple[Never, Never] to Never,
                # mypy doesn't. Hmm.
                typing.assert_never(unreachable)

    def _do_release(self):
        self._connec._release(self._id)
        self._status = _SavepointStatus.DEAD

    def _do_rollback_before(self):
        self._connec._rollback_to_before(self._id)
        self._status = _SavepointStatus.DEAD

    def _do_rollback_after(self):
        # connection will raise the rror if this isn't the innermost
        self._connec._rollback_to_after(self._id)

    @_must_be_active
    def rollback_to_after_this_innermost(self):
        """Roll back this, the innermost savepoint,
        to after it started.
        This will undo all changes since then.

        Note that this doesn't "re-run" the context manager block,
        it merely discards previous changes that happened within.

        .. warning::
            Using this on anything but the innermost savepoint
            is incorrect and will raise an
            :exc:`IncorrectTransactionControlError`.

            If you wish to use this on an outer savepoint,
            see :meth:`ROLLBACK_TO_AFTER_NOT_INNERMOST`.
        """
        self._do_rollback_after()

    @_must_be_active
    def ROLLBACK_TO_AFTER_NOT_INNERMOST(self) -> Never:
        """Roll back this, an outer savepoint,
        to after it started.
        Specifically, this will raise an exception
        that is caught by the context manager one level
        within this one.

        Note that this doesn't "re-run" the context manager block,
        it merely discards previous changes that happened within.

        See the :mod:`module docs <coolqlite.transactions>`
        for details.

        .. warning::
            Using this on anything but an outer savepoint
            is incorrect and will raise an
            :exc:`IncorrectTransactionControlError`.

            If you wish to use this on an inner savepoint,
            see :meth:`rollback_to_after_this_innermost`.
        """
        raise _RollbackToAfterNotInnermost(self._id)

    @_must_be_active
    def ROLLBACK_TO_BEFORE(self) -> Never:
        """Roll back before this savepoint.
        Specifically, this will raise an exception
        that is caught by this context manager.

        See the :mod:`module docs <coolqlite.transactions>`
        for details.
        """
        raise _RollbackToBefore(self._id)

    @_must_be_active
    def RELEASE(self) -> Never:
        """Release this savepoint.
        If there are outer savepoints,
        this "merges" the changes into them.
        If this is the outermost savepoint,
        this commits changes to the database.
        Specifically, this will raise an exception
        that is caught by this context manager.

        See the :mod:`module docs <coolqlite.transactions>`
        for details.
        """
        raise _Release(self._id)


_OLD_TODOS = """
TODO: if each savepoint keeps a reference to its parent,
then it can "prime" them for receiving the exception.
if on or method call it doesn't, then that means something swallowed
it and that's an IncorrectError!

TODO: proper del behavior. should probably be an error if active?

TODO: don't forget to check for generic sql errors
when doing an operation, just in case there was an
`ON CONFLICT ROLLBACK`.
but in the methods that handle it.
or do we make note of it here
so we might guess that something is tainted?
eh would be error prone...
ughhhh what's worse or what's better.
for now: explicit is better than implicit.
if you don't deactivate these yourself,
that's an error.

TODO: can we somehow handle an integrity error ourselves?
we probably don't know if it's from an abort
or a rollback.
unless we check right after, right?
if they caught it, on the next savepoint-related call,
we can check.

TODO: clarify in docs that caps methods use exceptions.
"""

__all__ = [
    "AdvancedSavepoint",
    "TransactionControlException",
    "IncorrectTransactionControlError",
]



"""Helpers for configuring cattrs with our default
good-practices converters.
"""

from typing import get_args, overload
import cattrs
import cattrs.strategies
from datetype import (
    AwareDateTime,
    DateTime,
    NaiveDateTime,
    Date,
    AwareTime,
    NaiveTime,
    Time,
)

from coolqlite._common import SqliteType
from coolqlite.from_sql.datetime_ import (
    aware_dt_from_iso,
    aware_time_from_iso,
    date_from_iso,
    dt_from_iso,
    naive_dt_from_iso,
    naive_time_from_iso,
    time_from_iso,
)


def configure_converter_datetimes(conv: cattrs.BaseConverter):
    """Configure the given converter with structure hooks for
    the various datetype types.
    """
    conv.register_structure_hook(DateTime, dt_from_iso)
    conv.register_structure_hook(AwareDateTime, aware_dt_from_iso)
    conv.register_structure_hook(NaiveDateTime, naive_dt_from_iso)
    conv.register_structure_hook(NaiveDateTime | AwareDateTime, dt_from_iso)

    conv.register_structure_hook(Time, time_from_iso)
    conv.register_structure_hook(AwareTime, aware_time_from_iso)
    conv.register_structure_hook(NaiveTime, naive_time_from_iso)
    conv.register_structure_hook(NaiveTime | AwareTime, time_from_iso)

    conv.register_structure_hook(Date, date_from_iso)


# TODO: ints as floats
def _strict_primitive_hook(strict_type: type):
    def _hook(val, requested_type):
        # TODO: why does the example have it doing `requested_type` instead of int?
        # cuz then the return type is... well, _correct_, but...
        if not isinstance(val, requested_type):
            raise ValueError(f"{val!r} not an instance of {requested_type}")
        return val

    return _hook


def configure_strict_primitives(conv: cattrs.BaseConverter):
    for type_ in get_args(SqliteType):
        conv.register_structure_hook(type_, _strict_primitive_hook(type_))


def configure_union_passthrough(conv: cattrs.BaseConverter):
    cattrs.strategies.configure_union_passthrough(SqliteType, conv)


@overload
def configure_converter() -> cattrs.Converter: ...
@overload
def configure_converter[T: cattrs.BaseConverter = cattrs.BaseConverter](
    conv: T | None = None,
) -> T: ...
def configure_converter[T: cattrs.BaseConverter = cattrs.BaseConverter](
    conv: T | None = None,
) -> T | cattrs.Converter:
    """Configure a converter with our good-practices defaults.
    If not given, creates a new ``cattrs.Converter`` .
    """
    _conv = cattrs.Converter() if conv is None else conv
    # _conv.register_structure_hook(Union, )
    configure_strict_primitives(_conv)
    configure_union_passthrough(_conv)
    configure_converter_datetimes(_conv)
    return _conv
